#include <stdlib.h>
#include <inttypes.h>
#include <memory.h>
#include <time.h>
#include <stdio.h>

#include "lib/lib.c"
#include "lib/rect.c"
#include "lib/celllist.c"
#include "lib/randomarray.c"
#include "lib/golgrid.c"
#include "lib/deprecated.c"
#include "lib/objects.c"
#include "lib/gridmisc.c"
#include "lib/hashtable.c"
#include "lib/store.c"

#define GRID_SIZE 64

#define MAX_SLOW_GLIDERS 32
#define MAX_CELLS 1024
#define MAX_BYTE_SEQ_SIZE 2560

#define SLOW_MAX_GENS 128
#define SLOW_MAX_SIZE 30

#define SEED_MAX_GENS 160
#define SEED_LANE_ON -10
#define SEED_LANE_OFF 14
#define SEED_SAFE_PROGRESSION 5

static u64 test_cnt = 0;
static u64 new_seed_cnt = 0;


typedef struct
{
	GoLGrid *allowed_area;
	
	GoLGrid *start_elbow;
	RandomDataArray *rda;
	HashTable_void *known_patterns;
	HashTable_void *useful_patterns;
	s32 slow_max_gens;
	s32 slow_max_size;
	
	GoLGrid *base_pattern;
	GoLGrid *open_area;
	GoLGrid *fixed_cells;
	GoLGrid *wanted_pattern;
	s32 seed_lane_on;
	s32 seed_lane_off;
	s32 seed_safe_progression;
	s32 seed_max_gens;
	
	s32 sol_cnt;
} SearchEnv;


static __not_inline GoLGrid *alloc_grid (void)
{
	Rect grid_rect;
	Rect_make (&grid_rect, -(GRID_SIZE / 2), -(GRID_SIZE / 2), GRID_SIZE, GRID_SIZE);
	
	GoLGrid *gg = malloc (sizeof (GoLGrid));
	GoLGrid_create (gg, &grid_rect);
	
	return gg;
}

static __not_inline void mirror_glider (Glider *gl)
{
	s32 lane_offs = (-gl->lane - 1) - gl->lane;
	gl->timing += (2 * lane_offs);
	gl->lane += lane_offs;
}

static __not_inline void flip_diagonally_abs (const GoLGrid *in_gg, GoLGrid *out_gg)
{
	static int static_init = FALSE;
	static GoLGrid *zeroed;
	static GoLGrid *flipped;
	
	if (!static_init)
	{
		static_init = TRUE;
		zeroed = alloc_grid ();
		flipped = alloc_grid ();
	}
	
	Rect bb;
	GoLGrid_get_bounding_box (in_gg, &bb);
	
	GoLGrid_copy_unmatched_noinline (in_gg, zeroed, -(GRID_SIZE / 2) - bb.left_x, -(GRID_SIZE / 2) - bb.top_y);
	GoLGrid_flip_diagonally (zeroed, flipped);
	GoLGrid_copy_unmatched_noinline (flipped, out_gg, (GRID_SIZE / 2) + bb.top_y, (GRID_SIZE / 2) + bb.left_x);
}

static __not_inline int run_pattern_even_gens (const GoLGrid *start_pattern, const GoLGrid *allowed_area, s32 max_gens, GoLGrid *settled_pattern)
{
	static int static_init = FALSE;
	static GoLGrid *p_m2;
	static GoLGrid *p_m1;
	static GoLGrid *p_m0;
	
	if (!static_init)
	{
		static_init = TRUE;
		p_m2 = alloc_grid ();
		p_m1 = alloc_grid ();
		p_m0 = alloc_grid ();
	}
	
	GoLGrid_copy_noinline (start_pattern, p_m0);
	
	s32 gen = 0;
	while (TRUE)
	{
		if (GoLGrid_is_empty (p_m0))
			return FALSE;
		
		if (!GoLGrid_is_subset (p_m0, allowed_area))
			return FALSE;
		
		if (gen >= 2 && (gen % 2 == 0) && GoLGrid_is_equal_64_wide (p_m0, p_m2))
		{
			GoLGrid_copy_noinline (p_m0, settled_pattern);
			return TRUE;
		}
		
		if (gen >= max_gens)
			return FALSE;
		
		GoLGrid *temp = p_m2;
		p_m2 = p_m1;
		p_m1 = p_m0;
		p_m0 = temp;
		
		GoLGrid_evolve_64_wide (p_m1, p_m0);
		gen++;
	}
}

// Get info of how to shoot NW gliders at the pattern. Safe glider timing is ((2 * lane) - safe_progression) or higher
static __not_inline void get_lane_range_and_safe_progression (const GoLGrid *pattern, int is_p2, s32 *lane_on, s32 *lane_off, s32 *safe_progression)
{
	static int static_init = FALSE;
	static GoLGrid *envelope;
	
	if (!static_init)
	{
		static_init = TRUE;
		envelope = alloc_grid ();
	}
	
	if (is_p2)
		GoLGrid_evolve_noinline (pattern, envelope);
	else
		GoLGrid_clear_noinline (envelope);
	
	GoLGrid_or_noinline (envelope, pattern);
	
	u64 projection [2];
	GoLGrid_make_rightdown_projection_noinline (envelope, projection, 2);
	
	*lane_on = GoLGrid_get_rightup_pop_on (envelope, projection, 2) - 5;
	*lane_off = GoLGrid_get_rightup_pop_off (envelope, projection, 2) + 4;
	
	GoLGrid_make_rightup_projection_noinline (envelope, projection, 2);
	
	*safe_progression = -9 - (2 * GoLGrid_get_rightdown_pop_off (envelope, projection, 2));
}

static __not_inline void print_slow_gliders (const GoLGrid *start_elbow, Glider slow_gl [], s32 slow_gl_cnt)
{
	static int static_init = FALSE;
	static GoLGrid print_grid;
	
	if (!static_init)
	{
		static_init = TRUE;
		
		Rect r;
		Rect_make (&r, -64, -64, 1024, 1024);
		
		GoLGrid_create (&print_grid, &r);
	}
	
	GoLGrid_copy_unmatched_noinline (start_elbow, &print_grid, 0, 0);
	
	s32 glider_ix;
	for (glider_ix = 0; glider_ix < slow_gl_cnt; glider_ix++)
	{
		Glider gl;
		gl.dir = slow_gl [glider_ix].dir;
		gl.lane = slow_gl [glider_ix].lane;
		gl.timing = slow_gl [glider_ix].timing + 16 + 96 * glider_ix;
		
		if (glider_ix > 0)
			printf (" ");
		
		printf ("%c%d", (gl.timing & 1) ? 'O' : 'E', gl.lane);
		
		GoLGrid_or_glider (&print_grid, &gl);
	}
	
	printf ("\n");
	GoLGrid_print_life_history (&print_grid);
}

static __not_inline s32 test_slow_recipe (const GoLGrid *backdrop, const GoLGrid *fixed_cells, Glider slow_gl [], s32 slow_gl_cnt, s32 gens_per_glider, int is_mirrored, s32 offs_x, s32 offs_y)
{
	static int static_init = FALSE;
	static GoLGrid *p_m1;
	static GoLGrid *p_m0;
	
	if (!static_init)
	{
		static_init = TRUE;
		p_m1 = alloc_grid ();
		p_m0 = alloc_grid ();
	}
	
	GoLGrid_copy_noinline (backdrop, p_m0);
	
	s32 glider_ix;
	for (glider_ix = 0; glider_ix < slow_gl_cnt; glider_ix++)
	{
		Glider gl;
		gl.dir = slow_gl [glider_ix].dir;
		gl.lane = slow_gl [glider_ix].lane;
		gl.timing = slow_gl [glider_ix].timing;
		
		if (is_mirrored)
			mirror_glider (&gl);
		
		gl.lane += (offs_x - offs_y);
		gl.timing += 4 * offs_x;
		
		GoLGrid_set_generation (p_m0, 0);
		GoLGrid_or_glider (p_m0, &gl);
		
		s32 gen = 0;
		while (TRUE)
		{
			if (gen >= gens_per_glider)
				break;
			
			GoLGrid_copy_noinline (p_m0, p_m1);
			GoLGrid_evolve_noinline (p_m1, p_m0);
			gen++;
		}
		
		if (!GoLGrid_is_subset_noinline (fixed_cells, p_m0))
			return glider_ix;
	}
	
	return -1;
}

static __not_inline int test_pattern (const GoLGrid *start, SearchEnv *se, Glider slow_gl [], s32 slow_gl_cnt)
{
	static int static_init = FALSE;
	static GoLGrid *p_m2;
	static GoLGrid *p_m1;
	static GoLGrid *p_m0;
	
	if (!static_init)
	{
		static_init = TRUE;
		p_m2 = alloc_grid ();
		p_m1 = alloc_grid ();
		p_m0 = alloc_grid ();
	}
	
	GoLGrid_copy_noinline (start, p_m0);
	
	test_cnt++;
	
	if (test_cnt % 10000000 == 0)
		fprintf (stderr, "Testing pattern %dM, average per new seed = %d\n", (s32) (test_cnt / 1000000), (s32) (test_cnt / new_seed_cnt));
	
	s32 gen = 0;
	while (TRUE)
	{
		if (!GoLGrid_is_subset_64_wide (p_m0, se->allowed_area))
			return FALSE;
		
		if (!GoLGrid_is_subset_64_wide (se->fixed_cells, p_m0))
			return FALSE;
		
		if (gen >= 2 && GoLGrid_is_equal_64_wide (p_m0, p_m2))
		{
			if (GoLGrid_is_subset (se->wanted_pattern, p_m0))
			{
				se->sol_cnt++;
				printf ("Found a solution, population %d:\n", (s32) GoLGrid_get_population (p_m0));
				GoLGrid_print (start);
				GoLGrid_print_life_history (start);
				print_slow_gliders (se->start_elbow, slow_gl, slow_gl_cnt);
				return TRUE;
			}
			
			return FALSE;
		}
		
		if (gen >= se->seed_max_gens)
			return FALSE;
		
		GoLGrid *temp = p_m2;
		p_m2 = p_m1;
		p_m1 = p_m0;
		p_m0 = temp;
		
		GoLGrid_evolve_64_wide (p_m1, p_m0);
		gen++;
	}
}

static __not_inline int test_seed_phase (const GoLGrid *seed, SearchEnv *se, Glider slow_gl [], s32 slow_gl_cnt, int is_mirrored)
{
	static int static_init = FALSE;
	static GoLGrid *seed_envelope;
	static GoLGrid *seed_envelope_pos;
	static GoLGrid *seeded_pattern;
	static GoLGrid *start_pattern;
	static GoLGrid *verify_elbow;
	static GoLGrid *verify_start;
	
	if (!static_init)
	{
		static_init = TRUE;
		seed_envelope = alloc_grid ();
		seed_envelope_pos = alloc_grid ();
		seeded_pattern = alloc_grid ();
		start_pattern = alloc_grid ();
		verify_elbow = alloc_grid ();
		verify_start = alloc_grid ();
	}
	
	GoLGrid_evolve_noinline (seed, seed_envelope);
	GoLGrid_or_noinline (seed_envelope, seed);
	
	Rect area_bb;
	GoLGrid_get_bounding_box (se->open_area, &area_bb);
	
	Rect seed_bb;
	GoLGrid_get_bounding_box (seed_envelope, &seed_bb);
	
	int useful_seed = FALSE;
	s32 offs_x;
	s32 offs_y;
	
	for (offs_x = area_bb.left_x - seed_bb.left_x; offs_x < (area_bb.left_x + area_bb.width) - (seed_bb.left_x + seed_bb.width); offs_x++)
		for (offs_y = area_bb.top_y - seed_bb.top_y; offs_y < (area_bb.top_y + area_bb.height) - (seed_bb.top_y + seed_bb.height); offs_y++)
		{
			GoLGrid_copy_unmatched_noinline (seed_envelope, seed_envelope_pos, offs_x, offs_y);
			
			if (GoLGrid_is_subset (seed_envelope_pos, se->open_area))
			{
				GoLGrid_copy_unmatched_noinline (seed, seeded_pattern, offs_x, offs_y);
				GoLGrid_or_noinline (seeded_pattern, se->base_pattern);
				
				s32 lane_ix;
				for (lane_ix = se->seed_lane_on; lane_ix < se->seed_lane_off; lane_ix++)
				{
					GoLGrid_copy (seeded_pattern, start_pattern);
					
					Glider gl;
					gl.dir = 0;
					gl.lane = lane_ix;
					gl.timing = 2 * gl.lane - se->seed_safe_progression;
					GoLGrid_or_glider (start_pattern, &gl);
					
					int found = test_pattern (start_pattern, se, slow_gl, slow_gl_cnt);
					useful_seed |= found;
					
					if (found)
					{
						if (is_mirrored)
							flip_diagonally_abs (se->start_elbow, verify_elbow);
						else
							GoLGrid_copy_noinline (se->start_elbow, verify_elbow);
						
						GoLGrid_copy_unmatched_noinline (verify_elbow, verify_start, offs_x, offs_y);
						
						GoLGrid_or_noinline (verify_start, se->base_pattern);
						
						s32 failed_glider = test_slow_recipe (verify_start, se->fixed_cells, slow_gl, slow_gl_cnt, se->slow_max_gens, is_mirrored, offs_x, offs_y);
						if (failed_glider == -1)
							printf ("\n--- Solution works! ---\n\n");
						else
							printf ("Failed verification at glider %d\n", failed_glider + 1);
					}
				}
			}
		}
	
	return useful_seed;
}

static __not_inline int test_seed_in_place (const GoLGrid *seed, SearchEnv *se, Glider slow_gl [], s32 slow_gl_cnt)
{
	static int static_init = FALSE;
	static GoLGrid *mirrored_seed;
	static GoLGrid *evolved;
	
	if (!static_init)
	{
		static_init = TRUE;
		mirrored_seed = alloc_grid ();
		evolved = alloc_grid ();
	}
	
	int useful_seed = test_seed_phase (seed, se, slow_gl, slow_gl_cnt, FALSE);
	
	flip_diagonally_abs (seed, mirrored_seed);
	useful_seed |= test_seed_phase (mirrored_seed, se, slow_gl, slow_gl_cnt, TRUE);
	
	// Generation must be reset or the glider will be added in the wrong progression
	GoLGrid_evolve_noinline (seed, evolved);
	GoLGrid_set_generation (evolved, 0);
	
	if (!GoLGrid_is_equal (evolved, seed))
	{
		useful_seed |= test_seed_phase (evolved, se, slow_gl, slow_gl_cnt, FALSE);
		
		flip_diagonally_abs (evolved, mirrored_seed);
		useful_seed |= test_seed_phase (mirrored_seed, se, slow_gl, slow_gl_cnt, TRUE);
	}
	
	return useful_seed;
}

static __not_inline void store_pattern (const GoLGrid *pattern, Glider slow_gl [], s32 slow_gl_cnt, ByteSeqStore *bss)
{
	s32 byte_seq_ix = 0;
	u8 byte_seq [MAX_BYTE_SEQ_SIZE];
	
	s32 population = GoLGrid_get_population (pattern);
	
	byte_seq [byte_seq_ix++] = (u8) (population >> 8);
	byte_seq [byte_seq_ix++] = (u8) (population & 0xff);
	
	int first_cell = TRUE;
	s32 cell_x = 0;
	s32 cell_y = 0;
	while (TRUE)
	{
		if (!GoLGrid_find_next_on_cell (pattern, first_cell, &cell_x, &cell_y))
			break;
		
		first_cell = FALSE;
		byte_seq [byte_seq_ix++] = (u8) (s8) cell_x;
		byte_seq [byte_seq_ix++] = (u8) (s8) cell_y;
	}
	
	byte_seq [byte_seq_ix++] = (u8) slow_gl_cnt;
	
	s32 slow_gl_ix;
	for (slow_gl_ix = 0; slow_gl_ix < slow_gl_cnt; slow_gl_ix++)
	{
		byte_seq [byte_seq_ix++] = (u8) (s8) slow_gl [slow_gl_ix].lane;
		byte_seq [byte_seq_ix++] = (u8) (s8) slow_gl [slow_gl_ix].timing;
	}
	
	ByteSeqStore_store (bss, byte_seq, byte_seq_ix);
}

static __not_inline void unpack_stored_pattern (const u8 byte_seq [], GoLGrid *pattern, Glider slow_gl [], s32 *slow_gl_cnt)
{
	s32 byte_seq_ix = 0;
	
	s32 cell_cnt;
	cell_cnt = 256 * (s32) byte_seq [byte_seq_ix++];
	cell_cnt += (s32) byte_seq [byte_seq_ix++];
	
	GoLGrid_clear_noinline (pattern);
	
	s32 cell_ix;
	for (cell_ix = 0; cell_ix < cell_cnt; cell_ix++)
	{
		s32 cell_x = (s32) (s8) byte_seq [byte_seq_ix++];
		s32 cell_y = (s32) (s8) byte_seq [byte_seq_ix++];
		GoLGrid_set_cell_on (pattern, cell_x, cell_y);
	}
	
	*slow_gl_cnt = (s32) byte_seq [byte_seq_ix++];
	
	s32 slow_gl_ix;
	for (slow_gl_ix = 0; slow_gl_ix < *slow_gl_cnt; slow_gl_ix++)
	{
		slow_gl [slow_gl_ix].dir = 0;
		slow_gl [slow_gl_ix].lane = (s32) (s8) byte_seq [byte_seq_ix++];
		slow_gl [slow_gl_ix].timing = (s32) (s8) byte_seq [byte_seq_ix++];
	}
}

static __not_inline int is_in_hashtable (const GoLGrid *pattern, HashTable_void *ht, const RandomDataArray *rda)
{
	static int static_init = FALSE;
	static GoLGrid *zeroed;
	
	if (!static_init)
	{
		static_init = TRUE;
		zeroed = alloc_grid ();
	}
	
	Rect bb;
	GoLGrid_get_bounding_box (pattern, &bb);
	
	GoLGrid_copy_unmatched_noinline (pattern, zeroed, -(GRID_SIZE / 2) - bb.left_x, -(GRID_SIZE / 2) - bb.top_y);
	u64 hash = GoLGrid_get_hash_noinline (zeroed, rda);
	
	return HashTable_void_is_marked (ht, hash);
}

static __not_inline void store_in_hashtable (const GoLGrid *pattern, HashTable_void *ht, const RandomDataArray *rda)
{
	static int static_init = FALSE;
	static GoLGrid *zeroed;
	static GoLGrid *evolved;
	static GoLGrid *mirrored;
	
	if (!static_init)
	{
		static_init = TRUE;
		zeroed = alloc_grid ();
		evolved = alloc_grid ();
		mirrored = alloc_grid ();
	}
	
	Rect bb;
	GoLGrid_get_bounding_box (pattern, &bb);
	
	GoLGrid_copy_unmatched_noinline (pattern, zeroed, -(GRID_SIZE / 2) - bb.left_x, -(GRID_SIZE / 2) - bb.top_y);
	u64 hash = GoLGrid_get_hash_noinline (zeroed, rda);
	
	HashTable_void_mark (ht, hash, NULL);
	
	GoLGrid_flip_diagonally (zeroed, mirrored);
	HashTable_void_mark (ht, GoLGrid_get_hash_noinline (mirrored, rda), NULL);
	
	GoLGrid_evolve_noinline (pattern, evolved);
	GoLGrid_get_bounding_box (evolved, &bb);
	GoLGrid_copy_unmatched_noinline (evolved, zeroed, -(GRID_SIZE / 2) - bb.left_x, -(GRID_SIZE / 2) - bb.top_y);
	HashTable_void_mark (ht, GoLGrid_get_hash_noinline (zeroed, rda), NULL);
	
	GoLGrid_flip_diagonally (zeroed, mirrored);
	HashTable_void_mark (ht, GoLGrid_get_hash_noinline (mirrored, rda), NULL);
}

static __not_inline void process_seed (const GoLGrid *seed, SearchEnv *se, Glider slow_gl [], s32 in_slow_gl_cnt, ByteSeqStore *bss)
{
	static int static_init = FALSE;
	static GoLGrid *evolved;
	static GoLGrid *start_pattern;
	static GoLGrid *settled_pattern;
	
	if (!static_init)
	{
		static_init = TRUE;
		evolved = alloc_grid ();
		start_pattern = alloc_grid ();
		settled_pattern = alloc_grid ();
	}
	
	GoLGrid_evolve_noinline (seed, evolved);
	s32 phase_off = (GoLGrid_is_equal (evolved, seed) ? 1 : 2);
	
	s32 lane_on;
	s32 lane_off;
	s32 safe_progression;
	
	get_lane_range_and_safe_progression (seed, (phase_off == 2), &lane_on, &lane_off, &safe_progression);
	
	Glider gl;
	gl.dir = 0;
	
	s32 phase_ix;
	for (gl.lane = lane_on; gl.lane < lane_off; gl.lane++)
		for (phase_ix = 0; phase_ix < phase_off; phase_ix++)
		{
			GoLGrid_copy_noinline (seed, start_pattern);
			
			gl.timing = 2 * gl.lane - safe_progression + phase_ix;
			GoLGrid_or_glider (start_pattern, &gl);
			
			if (run_pattern_even_gens (start_pattern, se->allowed_area, se->slow_max_gens, settled_pattern))
			{
				Rect bb;
				GoLGrid_get_bounding_box (settled_pattern, &bb);
				
				if (bb.width + bb.height <= se->slow_max_size)
				{
					int was_known = is_in_hashtable (settled_pattern, se->known_patterns, se->rda);
					int known_useful = is_in_hashtable (settled_pattern, se->useful_patterns, se->rda);
					
					if (!was_known || known_useful)
					{
						if (!was_known)
							store_in_hashtable (settled_pattern, se->known_patterns, se->rda);
						
						slow_gl [in_slow_gl_cnt].dir = 0;
						slow_gl [in_slow_gl_cnt].lane = gl.lane;
						slow_gl [in_slow_gl_cnt].timing = gl.timing;
						
						store_pattern (settled_pattern, slow_gl, in_slow_gl_cnt + 1, bss);
						
						new_seed_cnt++;
						
						GoLGrid_set_generation (settled_pattern, 0);
						if (test_seed_in_place (settled_pattern, se, slow_gl, in_slow_gl_cnt + 1))
							if (!known_useful)
								store_in_hashtable (settled_pattern, se->useful_patterns, se->rda);
					}
				}
			}
		}
}

static void run_store (const ByteSeqStore *in_bss, SearchEnv *se, ByteSeqStore *out_bss)
{
	static int static_init = FALSE;
	static GoLGrid *start_pattern;
	
	if (!static_init)
	{
		static_init = TRUE;
		start_pattern = alloc_grid ();
	}
	
	ByteSeqStoreNode *bss_node;
	s32 node_data_offset;
	ByteSeqStore_start_get_iteration (in_bss, &bss_node, &node_data_offset);
	
	s32 store_ix = 0;	
	while (TRUE)
	{
		u8 byte_seq [MAX_BYTE_SEQ_SIZE];
		if (!ByteSeqStore_get_next (in_bss, &bss_node, &node_data_offset, byte_seq, MAX_BYTE_SEQ_SIZE, NULL))
			break;
		
		Glider slow_gl [MAX_SLOW_GLIDERS];
		s32 slow_gl_cnt;
		unpack_stored_pattern (byte_seq, start_pattern, slow_gl, &slow_gl_cnt);
		
		if (store_ix % 10000 == 0)
			fprintf (stderr, "Running pattern %d, solutions found: %d\n", store_ix, se->sol_cnt);
		
		process_seed (start_pattern, se, slow_gl, slow_gl_cnt, out_bss);
		store_ix++;
	}
}

int main ()
{
	SearchEnv se;
	
	se.allowed_area = alloc_grid ();
	s32 y;
	s32 x;
	for (y = -31; y < 31; y++)
		for (x = -31; x < 31; x++)
			GoLGrid_set_cell_on (se.allowed_area, x, y);
	
	se.start_elbow = alloc_grid ();
	
	RandomDataArray rda;
	se.rda = &rda;
	RandomDataArray_create (se.rda, GRID_SIZE * GRID_SIZE / 64);
	
	HashTable_void known_patterns;
	se.known_patterns = &known_patterns;
	HashTable_void_create (se.known_patterns, 256, 0.7, 0.9);
	
	HashTable_void useful_patterns;
	se.useful_patterns = &useful_patterns;
	HashTable_void_create (se.useful_patterns, 256, 0.7, 0.9);
	
	se.slow_max_gens = SLOW_MAX_GENS;
	se.slow_max_size = SLOW_MAX_SIZE;
	
	se.base_pattern = alloc_grid ();
	se.open_area = alloc_grid ();
	se.fixed_cells = alloc_grid ();
	se.wanted_pattern = alloc_grid ();
	
	se.seed_lane_on = SEED_LANE_ON;
	se.seed_lane_off = SEED_LANE_OFF;
	se.seed_safe_progression = SEED_SAFE_PROGRESSION;
	se.seed_max_gens = SEED_MAX_GENS;
	
	se.sol_cnt = 0;
	
	GoLGrid_parse_life_history_simple (".A$.A$.A!", -8, -8, se.start_elbow);

// Gliders from SE, lane_on = -10, lane_off = 19, safe_progression = -17
//	const char *start_spec = "8.C$4.2C.C.A6.6D$3.C.C.C.A5.8D$C2.C.C.C.A.A2.10D$4C.2C3.2A2.11D$4.C9.11D$2.C.C3.D4.12D$2.2C3.18D$6.19D$5.20D$4.21D$4.21D$4.20D$4.19D$5.17D$6.15D$7.13D$8.11D!";
//	const char *wanted_spec = "8.A$4.2A.A.A$3.A.A.A.A$A2.A.A.A.A$4A.2A2.A$4.A4.A$2.A.A$2.2A!";
	
// Gliders from SW, lane_on = -10, lane_off = 15, safe_progression = -13
//	const char *start_spec = "4.8D$3.10D$2.12D$.14D$.15D$.16D$.17D$.17D$.17D$2.16D$3.15D$6.12D$7.11D$3.2A2.11D$4.A2.11D$.3A3.11D$C5.12D$.3C3.10D$4.C3.8D$.4C4.6D$.C3.2C3.4D$2.3C2.C$4.C.2C$4.C$3.2C!";
//	const char *wanted_spec = "15$.5A$A$.3A$4.A$.4A$.A3.2A$2.3A2.A$4.A.2A$4.A$3.2A!";
	
// Gliders from NE, lane_on = -18, lane_off = 4, safe_progression = -17
//	const char *start_spec = "15.2C$15.C$12.2C.C$12.C2.3C$6.4D3.2C3.C$5.6D4.4C$4.8D3.C$3.10D3.3C$2.12D5.C$.12D3.3A$13D2.A$13D2.2A$13D$14D$17D$18D$18D$.17D$2.16D$3.14D$4.12D$5.10D$6.8D$7.6D$8.4D!";
//	const char *wanted_spec = "15.2A$15.A$12.2A.A$12.A2.3A$13.2A3.A$15.4A$15.A$16.3A$19.A$14.5A!";
	
// Gliders from NW, lane_on = -20, lane_off = -5, safe_progression = -1
//	const char *start_spec = "7.8D$6.10D$5.12D$4.14D$3.16D$2.18D$.19D$20D$20D$20D$19D3.2C$13D4.D3.C.C$12D9.C$12D2.2A3.2C.4C$12D2.A.A.C.C.C2.C$12D4.A.C.C.C$.11D4.A.C.2C$2.9D6.C$3.7D!";
//	const char *wanted_spec = "10$22.2A$21.A.A$16.A4.A$16.A2.2A.4A$16.A.A.A.A2.A$16.A.A.A.A$16.A.A.2A$17.A!";

// Hard eater in Snark spiral, lane_on = -7, lane_off = 7, safe_progression = -31
//	const char *start_spec = "10.C$.2C6.C.C$.C.C5.C.C$3.C6.C$3.2C17.2C$22.2C2$5.16D$4.18D$3.20D$3.21D$3.22D$3.22D$3.22D$3.22D$3.22D$4.21D$5.20D$6.18D$7.16D$12.10D$13.8D$9.2C2.7D$8.C.C3.5D$2.2C4.C6.3D$C2.C2.2C.4C$2C.C.C.C.C2.C$3.C.C.C.C$3.C.C.2C$4.C!";
//	const char *wanted_spec = "10.A$.2A6.A.A$.A.A5.A.A$3.A6.A$3.2A17.2A$22.2A12$5.2A$6.A$3.3A$3.A2$9.2A$8.A.A$2.2A4.A$A2.A2.2A.4A$2A.A.A.A.A2.A$3.A.A.A.A$3.A.A.2A$4.A!";

// Add tail, lane_on = -10, lane_off = 14, safe_progression = 5
	const char *start_spec = "3.9D$2.11D$.13D$.14D$.14D$.14D$.14D$.14D$.14D$3.12D$4.11D$2C3.9D$C.A2.8D$.A3.7D!";
	const char *wanted_spec = "9$3.2A$3.A$2A.A$A.A$.A!";

	GoLGrid_parse_life_history (start_spec, -16, -16, se.base_pattern, se.open_area, NULL, NULL, NULL, NULL);
	GoLGrid_and_noinline (se.base_pattern, se.open_area, se.fixed_cells);
	GoLGrid_subtract_noinline (se.open_area, se.fixed_cells);
	
	GoLGrid_parse_life_history_simple (wanted_spec, -16, -16, se.wanted_pattern);
	GoLGrid_print_life_history_full (stderr, NULL, se.base_pattern, se.wanted_pattern, se.open_area, NULL);
	
	GoLGrid *show = alloc_grid ();
	GoLGrid_copy_noinline (se.base_pattern, show);
	Glider gl;
	gl.dir = 0;
	gl.lane = SEED_LANE_ON;
	gl.timing = 2 * gl.lane - se.seed_safe_progression;
	GoLGrid_or_glider (show, &gl);
	gl.lane = SEED_LANE_OFF - 1;
	gl.timing = 2 * gl.lane - se.seed_safe_progression;
	GoLGrid_or_glider (show, &gl);
	
	fprintf (stderr, "\n--- Please verify glider lanes and insertion timing:\n");
	GoLGrid_print_full (show, NULL, stderr);
	
	ByteSeqStore bss [MAX_SLOW_GLIDERS];
	s32 bss_ix;
	for (bss_ix = 0; bss_ix < MAX_SLOW_GLIDERS; bss_ix++)
		ByteSeqStore_create (&bss [bss_ix], 16384);
	
	Glider slow_gl [0];
	store_pattern (se.start_elbow, slow_gl, 0, &bss [0]);
	
	s32 slow_gl_cnt;
	for (slow_gl_cnt = 1; slow_gl_cnt <= 16; slow_gl_cnt++)
	{
		run_store (&bss [slow_gl_cnt - 1], &se, &bss [slow_gl_cnt]);
		fprintf (stderr, "Patterns in store for %d slow gliders: %d\n", slow_gl_cnt, (s32) bss [slow_gl_cnt].seq_count);
	}
	
	return EXIT_SUCCESS;
}
