#include <stdlib.h>
#include <inttypes.h>
#include <memory.h>
#include <time.h>
#include <stdio.h>

#include "lib/lib.c"
#include "lib/rect.c"
#include "lib/celllist.c"
#include "lib/randomarray.c"
#include "lib/golgrid.c"
#include "lib/deprecated.c"
#include "lib/objects.c"
#include "lib/gridmisc.c"

#define GRID_SIZE 64
#define OBJ_TYPE_CNT 41
#define MAX_OBJ_CELLS 8
#define MAX_POSS_OBJ 16384

#define MAX_GENS 128
#define OBJ_TYPE_USED_OFF 17
#define ADDED_OBJ_CNT 3
#define LANE_ON 14
#define LANE_OFF 26
#define P2_PHASE 0
#define GLIDER_DIR 2
#define TIMING_OFFSET 60
#define TIMING_FACTOR (-4)

#define TEST_ACCEPT_GEN_ON 40
#define ACCEPT_GEN_WAIT 16
#define MAX_COST 44


static ObjCellList_Cell obj_type_cells [OBJ_TYPE_CNT] [MAX_OBJ_CELLS];
static ObjCellList obj_type [OBJ_TYPE_CNT];

typedef struct
{
	s32 obj_type_ix;
	s32 x;
	s32 y;
	s32 cost;
} PossObj;

//                                block,     hive,             loaf,             boat,  blinker,     ship, pond,  tub,         longboat,  biblock,                             toad,           beacon,    barge,            mango
static const s32 obj_type_cost [] = {10,   12, 12,   14, 14, 14, 14,   14, 14, 14, 14,   11, 11,   14, 14,   14,   14,   18, 18, 18, 18,   18, 18,   20, 20, 20, 20, 20, 20, 20, 20,   21, 21, 21, 21,   20, 20,   22, 22, 22, 22};

static s32 poss_obj_cnt = 0;
static PossObj poss_obj [MAX_POSS_OBJ];


static __not_inline GoLGrid *alloc_grid (void)
{
	Rect grid_rect;
	Rect_make (&grid_rect, -(GRID_SIZE / 2), -(GRID_SIZE / 2), GRID_SIZE, GRID_SIZE);
	
	GoLGrid *gg = malloc (sizeof (GoLGrid));
	GoLGrid_create (gg, &grid_rect);
	
	return gg;
}

static __not_inline void copy_4_way_symmetry (s32 obj_type_ix)
{
	ObjCellList_copy (&obj_type [obj_type_ix], &obj_type [obj_type_ix + 1]);
	ObjCellList_copy (&obj_type [obj_type_ix], &obj_type [obj_type_ix + 3]);
	
	ObjCellList_flip_horizontally (&obj_type [obj_type_ix + 1]);
	
	ObjCellList_copy (&obj_type [obj_type_ix + 1], &obj_type [obj_type_ix + 2]);
	ObjCellList_flip_diagonally (&obj_type [obj_type_ix + 2]);
	
	ObjCellList_copy (&obj_type [obj_type_ix + 2], &obj_type [obj_type_ix + 3]);
	ObjCellList_flip_horizontally (&obj_type [obj_type_ix + 3]);
}

static __not_inline void make_obj_types (void)
{
	const char *block = "2A$2A!";
	const char *hive = ".A$A.A$A.A$.A!";
	const char *loaf = ".2A$A2.A$A.A$.A!";
	const char *boat = "2A$A.A$.A!";
	const char *blinker = "A$A$A!";
	const char *ship = "2A$A.A$.2A!";
	const char *pond = ".2A$A2.A$A2.A$.2A!";
	const char *tub = ".A$A.A$.A!";
	const char *longboat = "2A$A.A$.A.A$2.A!";
	const char *biblock = "2A$2A2$2A$2A!";
	const char *toad = "A$2A$2A$.A!";
	const char *beacon = "2A$A$3.A$2.2A!";
	const char *barge = ".A$A.A$.A.A$2.A!";
	const char *mango = ".A$A.A$A2.A$.A.A$2.A!";
	
	s32 obj_type_ix;
	for (obj_type_ix = 0; obj_type_ix < OBJ_TYPE_CNT; obj_type_ix++)
		ObjCellList_make_empty (&obj_type [obj_type_ix], obj_type_cells [obj_type_ix], MAX_OBJ_CELLS);
	
	ObjCellList_parse_rle (block, &obj_type [0]);
	
	ObjCellList_parse_rle (hive, &obj_type [1]);
	ObjCellList_copy (&obj_type [1], &obj_type [2]);
	ObjCellList_flip_diagonally (&obj_type [2]);
	
	ObjCellList_parse_rle (loaf, &obj_type [3]);
	copy_4_way_symmetry (3);
	
	ObjCellList_parse_rle (boat, &obj_type [7]);
	copy_4_way_symmetry (7);
	
	ObjCellList_parse_rle (blinker, &obj_type [11]);
	ObjCellList_evolve_slow (&obj_type [11], &obj_type [12]);
	
	ObjCellList_parse_rle (ship, &obj_type [13]);
	ObjCellList_copy (&obj_type [13], &obj_type [14]);
	ObjCellList_flip_horizontally (&obj_type [14]);

	ObjCellList_parse_rle (pond, &obj_type [15]);
	ObjCellList_parse_rle (tub, &obj_type [16]);
	
	ObjCellList_parse_rle (longboat, &obj_type [17]);
	copy_4_way_symmetry (17);
	
	ObjCellList_parse_rle (biblock, &obj_type [21]);
	ObjCellList_copy (&obj_type [21], &obj_type [22]);
	ObjCellList_flip_diagonally (&obj_type [22]);
	
	ObjCellList_parse_rle (toad, &obj_type [23]);
	copy_4_way_symmetry (23);
	ObjCellList_evolve_slow (&obj_type [23], &obj_type [27]);
	copy_4_way_symmetry (27);
	
	ObjCellList_parse_rle (beacon, &obj_type [31]);
	ObjCellList_copy (&obj_type [31], &obj_type [32]);
	ObjCellList_flip_horizontally (&obj_type [32]);
	ObjCellList_evolve_slow (&obj_type [31], &obj_type [33]);
	ObjCellList_copy (&obj_type [33], &obj_type [34]);
	ObjCellList_flip_horizontally (&obj_type [34]);
	
	ObjCellList_parse_rle (barge, &obj_type [35]);
	ObjCellList_copy (&obj_type [35], &obj_type [36]);
	ObjCellList_flip_horizontally (&obj_type [36]);
	
	ObjCellList_parse_rle (mango, &obj_type [37]);
	copy_4_way_symmetry (37);
}

static __not_inline void add_poss_objs (const GoLGrid *open_cells)
{
	static int static_init = FALSE;
	static GoLGrid *obj;
	static GoLGrid *obj_ev;
	
	if (!static_init)
	{
		static_init = TRUE;
		obj = alloc_grid ();
		obj_ev = alloc_grid ();
	}
	
	Rect bb;
	GoLGrid_get_bounding_box (open_cells, &bb);
	
	s32 y;
	s32 x;
	s32 obj_type_ix;
	for (y = bb.top_y; y < bb.top_y + bb.height; y++)
		for (x = bb.left_x; x < bb.left_x + bb.width; x++)
			for (obj_type_ix = 0; obj_type_ix < OBJ_TYPE_USED_OFF; obj_type_ix++)
			{
				GoLGrid_clear_noinline (obj);
				GoLGrid_or_obj_cell_list_noinline (obj, &obj_type [obj_type_ix], x, y);
				GoLGrid_evolve_noinline (obj, obj_ev);
				GoLGrid_or_noinline (obj, obj_ev);
				
				if (GoLGrid_is_subset_noinline (obj, open_cells))
				{
					if (poss_obj_cnt >= MAX_POSS_OBJ)
					{
						fprintf (stderr, "Too many possible objects\n");
						exit (0);
					}
					
					poss_obj [poss_obj_cnt].obj_type_ix = obj_type_ix;
					poss_obj [poss_obj_cnt].x = x;
					poss_obj [poss_obj_cnt].y = y;
					poss_obj_cnt++;
					
				}
			}
	
	fprintf (stderr, "Added %d possible objects\n", poss_obj_cnt);
}

static __force_inline int check_object_list (s32 poss_obj_ix [], s32 obj_cnt)
{
	s32 cost = 0;
	s32 obj_ix;
	for (obj_ix = 0; obj_ix < obj_cnt; obj_ix++)
	{
		cost += obj_type_cost [poss_obj [poss_obj_ix [obj_ix]].obj_type_ix];
		if (cost > MAX_COST)
			return obj_ix;
	}
	
	return -1;
}

static __not_inline int or_object_list (GoLGrid *base, s32 poss_obj_ix [], s32 obj_cnt)
{
	static int static_init = FALSE;
	static GoLGrid *new_obj;
	static GoLGrid *bleed_temp;
	static GoLGrid *bleed;
	
	if (!static_init)
	{
		static_init = TRUE;
		new_obj = alloc_grid ();
		bleed_temp = alloc_grid ();
		bleed = alloc_grid ();
	}
	
	GoLGrid_or_obj_cell_list (base, &obj_type [poss_obj [poss_obj_ix [0]].obj_type_ix], poss_obj [poss_obj_ix [0]].x, poss_obj [poss_obj_ix [0]].y);
	
	s32 obj_ix;
	for (obj_ix = 1; obj_ix < obj_cnt; obj_ix++)
	{
		GoLGrid_clear (new_obj);
		GoLGrid_or_obj_cell_list (new_obj, &obj_type [poss_obj [poss_obj_ix [obj_ix]].obj_type_ix], poss_obj [poss_obj_ix [obj_ix]].x, poss_obj [poss_obj_ix [obj_ix]].y);
		GoLGrid_bleed_8 (new_obj, bleed_temp);
		GoLGrid_bleed_4 (bleed_temp, bleed);
		
		if (!GoLGrid_are_disjoint (bleed, base))
			return obj_ix;
		
		GoLGrid_or (base, new_obj);
	}
	
	return -1;
}

static __not_inline int test_pattern (const GoLGrid *start, const GoLGrid *fixed_cells, const GoLGrid *end)
{
	static int static_init = FALSE;
	static GoLGrid *p_m2;
	static GoLGrid *p_m1;
	static GoLGrid *p_m0;
	
	if (!static_init)
	{
		static_init = TRUE;
		p_m2 = alloc_grid ();
		p_m1 = alloc_grid ();
		p_m0 = alloc_grid ();
	}
	
	GoLGrid_copy_noinline (start, p_m0);
	
	s32 gen = 0;
	s32 accept_gen_cnt = 0;
	while (TRUE)
	{
		if (gen >= TEST_ACCEPT_GEN_ON)
		{
			if (GoLGrid_is_subset (end, p_m0))
				accept_gen_cnt++;
			else
				accept_gen_cnt = 0;
			
			if (accept_gen_cnt >= ACCEPT_GEN_WAIT)
			{
				printf ("Found a solution:\n");
				GoLGrid_print (start);
				GoLGrid_print_life_history (start);
				return TRUE;
			}
		}
		
		if (gen >= 2 && GoLGrid_is_equal_64_wide (p_m0, p_m2) && accept_gen_cnt == 0)
			if (!GoLGrid_is_subset_noinline (end, p_m0))
				return FALSE;
		
		if (!GoLGrid_is_subset_64_wide (fixed_cells, p_m0))
			return FALSE;
		
		if (gen >= MAX_GENS)
			return FALSE;
		
		GoLGrid *temp = p_m2;
		p_m2 = p_m1;
		p_m1 = p_m0;
		p_m0 = temp;
		
		GoLGrid_evolve_64_wide (p_m1, p_m0);
		gen++;
	}
}

static __not_inline void init_setup (s32 obj_list [], s32 obj_list_cnt)
{
	s32 obj_ix;
	for (obj_ix = 0; obj_ix < obj_list_cnt; obj_ix++)
		obj_list [obj_ix] = obj_ix;
}

static __force_inline int next_setup (s32 obj_list [], s32 obj_ix_cnt, s32 failed_obj_ix, s32 poss_obj_cnt)
{
	s32 first_incr_ix = obj_ix_cnt - 1;
	if (failed_obj_ix != -1)
		first_incr_ix = failed_obj_ix;
	
	s32 incr_ix;
	s32 reset_ix;
	for (incr_ix = first_incr_ix; incr_ix >= 0; incr_ix--)
	{
		obj_list [incr_ix]++;
		for (reset_ix = incr_ix + 1; reset_ix < obj_ix_cnt; reset_ix++)
			obj_list [reset_ix] = obj_list [reset_ix - 1] + 1;
		
		if (obj_list [obj_ix_cnt - 1] < poss_obj_cnt)
			return TRUE;
	}
	
	return FALSE;
}

int main ()
{
	GoLGrid *base = alloc_grid ();
	GoLGrid *open_cells = alloc_grid ();
	GoLGrid *fixed_cells = alloc_grid ();
	GoLGrid *start = alloc_grid ();
	GoLGrid *end = alloc_grid ();
	GoLGrid *show = alloc_grid ();
	
// Snark conversion (LANE_ON 14, LANE_OFF 26, GLIDER_DIR 2, TIMING_OFFSET 60, TIMING_FACTOR (-4))
	const char *start_spec = "8.C$4.2C.C.A6.6D$3.C.C.C.A5.8D$C2.C.C.C.A.A2.10D$4C.2C3.2A2.11D$4.C9.12D$2.C.C8.13D$2.2C3.19D$6.20D$6.20D$6.20D$6.20D$6.20D$7.18D$8.16D$9.14D$10.12D$11.10D$12.8D!";
	const char *wanted_spec = "8.A$4.2A.A.A$3.A.A.A.A$A2.A.A.A.A$4A.2A2.A$4.A4.A$2.A.A$2.2A!";
	
//	Hard eater in Snark spiral (LANE_ON -8, LANE_OFF 12, GLIDER_DIR 0, TIMING_OFFSET 36, TIMING_FACTOR (+2))
//	const char *start_spec = "10.C$.2C6.C.C$.C.C5.C.C$3.C6.C$3.2C17.2C$22.2C2$7.9D$6.11D$5.13D$4.15D$3.17D$3.18D$3.18D$3.18D$4.17D$5.16D$6.14D$7.12D$8.10D$12.5D2$9.2C$8.C.C$2.2C4.C$C2.C2.2C.4C$2C.C.C.C.C2.C$3.C.C.C.C$3.C.C.2C$4.C!";
//	const char *wanted_spec = "10.A$.2A6.A.A$.A.A5.A.A$3.A6.A$3.2A17.2A$22.2A12$5.2A$6.A$3.3A$3.A2$9.2A$8.A.A$2.2A4.A$A2.A2.2A.4A$2A.A.A.A.A2.A$3.A.A.A.A$3.A.A.2A$4.A!";
	
//	Hard eater in Snark spiral, glider shifter (LANE_ON -10, LANE_OFF 8, GLIDER_DIR 0, TIMING_OFFSET 66, TIMING_FACTOR (+2))
//	const char *start_spec = "7.2A$6.A2.A$6.A2.A$7.2A$3.A$3.A$3.A5.2A$8.A2.A$8.A2.A$9.2A$.A$A.A$A.A6.2A$.A7.A.A.2C$11.A.C4.2C$11.C.C4.2C10.2C$12.C17.2C2$14.8D$9.14D$8.16D$8.17D$9.17D$10.16D$11.15D$12.14D$16.9D$17.8D$13.2C2.7D$14.C2.6D$11.3C3.5D$11.C2$17.2C$16.C.C$17.C!";
//	const char *wanted_spec = "11$10.2A$8.A2.A$8.2A.A.2A3.2A$11.A.A4.2A$11.A.A16.2A$12.A17.2A3$16.2A$16.2A8$13.2A$14.A$11.3A$11.A2$17.2A$16.A.A$17.A!";
	
	GoLGrid_parse_life_history (start_spec, -12, -12, base, open_cells, NULL, NULL, NULL, NULL);
	GoLGrid_and_noinline (base, open_cells, fixed_cells);
	GoLGrid_subtract_noinline (open_cells, fixed_cells);
	
	GoLGrid_parse_life_history_simple (wanted_spec, -12, -12, end);
	GoLGrid_print_life_history_full (stderr, NULL, base, end, NULL, NULL);
	
	GoLGrid_copy_noinline (base, show);
	Glider gl;
	gl.dir = GLIDER_DIR;
	gl.lane = LANE_ON;
	gl.timing = TIMING_OFFSET + (TIMING_FACTOR * LANE_ON);
	gl.timing += P2_PHASE;
	GoLGrid_or_glider (show, &gl);
	gl.lane = LANE_OFF - 1;
	gl.timing = TIMING_OFFSET + (TIMING_FACTOR * (LANE_OFF - 1));
	gl.timing += P2_PHASE;
	GoLGrid_or_glider (show, &gl);
	
	fprintf (stderr, "\n--- Please verify glider lanes and insertion timing:\n");
	GoLGrid_print_full (show, NULL, stderr);
	
	make_obj_types ();
	add_poss_objs (open_cells);
	
	u64 setup_cnt = 0;
	u64 test_cnt = 0;
	s32 solution_cnt = 0;
	
	s32 lane;
	for (lane = LANE_ON; lane < LANE_OFF; lane++)
	{
		fprintf (stderr, "Lane = %d\n", lane);
		
		gl.dir = GLIDER_DIR;
		gl.lane = lane;
		gl.timing = TIMING_OFFSET + (TIMING_FACTOR * lane);
		
		gl.timing += P2_PHASE;
		
		s32 obj_ix [ADDED_OBJ_CNT];
		init_setup (obj_ix, ADDED_OBJ_CNT);
		
		while (TRUE)
		{
			setup_cnt++;
			if ((setup_cnt & 0x00ffffffu) == 0)
				fprintf (stderr, "Setups: %dM, tested: %dM, first index: %d, solutions found: %d\n", (s32) (setup_cnt / 1000000), (s32) (test_cnt / 1000000), obj_ix [0], solution_cnt);
			
			s32 failing_obj_ix = check_object_list (obj_ix, ADDED_OBJ_CNT);
			
			if (failing_obj_ix == -1)
			{
				GoLGrid_copy_noinline (base, start);
				failing_obj_ix = or_object_list (start, obj_ix, ADDED_OBJ_CNT);
			}
			
			if (failing_obj_ix == -1)
			{
				if (!GoLGrid_or_glider (start, &gl))
					fprintf (stderr, "Attemped to place glider outside of grid\n");
				
				if (test_pattern (start, fixed_cells, end))
					solution_cnt++;
				
				test_cnt++;
			}
			
			if (!next_setup (obj_ix, ADDED_OBJ_CNT, failing_obj_ix, poss_obj_cnt))
				break;
		}
	}
	
	fprintf (stderr, "Done, %d solutions found\n", solution_cnt);
	return EXIT_SUCCESS;
}
